open Location;;

(* syntactic structure of expressions and programs *)

type location = int * int;;

type lpat = {pdesc : pat;
             ploc : location}

and

     pat = Nil 
         | Any
         | Cons of lpat * lpat
         | In of int
         | Id of string
;;

type ltypeexpr = { tdesc : ztypeexpr;
                   tloc : location}

and

     ztypeexpr = ZTypeVar of string
               | ZTypeConstr of ltypeexpr * string
               | ZTypeConst of string
               | ZTypeArrow of ltypeexpr * ltypeexpr
;;

type lexpr = {desc : zexpr;
              loc : location}

and

     zexpr = ZFunc of (lpat * lexpr) list
           | ZVar of string
           | ZInt of int
           | ZApply of lexpr * lexpr
           | ZAffectin of string * lexpr * lexpr
           | ZCond of lexpr * lexpr * lexpr
           | ZRec of string * lexpr
           | ZOp of string * lexpr * lexpr
           | ZMatch of lexpr * (lpat * lexpr) list
           | ZType of lexpr * ltypeexpr
;;

type zprogelem = ZAffect of string * lexpr
               | ZExpr of lexpr
;;

type sexpr = Cond of sexpr * sexpr * sexpr
           | Rec of string * sexpr
           | Op of string * sexpr * sexpr
           | F of string * sexpr
           | Func of string * sexpr
           | Var of string
           | Int of int
           | Apply of sexpr * sexpr
           | Affectin of string * sexpr * sexpr
;;


type sprogelem = Affect of string * sexpr
               | Expr of sexpr
;;

type param_t = PStr of string
             | PId of string
             | PInt of int
;;

type top_phrase = Command of string * param_t list
                | Prog of zprogelem
;;

let get_cur_loc() = (Parsing.symbol_start(), Parsing.symbol_end());;

exception Bad_end;;

let make_pat p = { pdesc = p;
                   ploc = get_cur_loc() }
;; 

let make_type t = { tdesc = t;
                    tloc = get_cur_loc() }
;;

let make_expr e = {desc = e;
                   loc = get_cur_loc() }
;;

let make_op op e1 e2 =
    make_expr (ZOp(op,e1,e2));;

(* flag to activate or not typing *)
let types = ref true;;

(* latex output on running or not ? *)
let latex = ref false;;

(* counting operations or not ? *)
let count = ref true;;

(* number of evaluations *)
let repeat = ref 1;;

(* machine type *)
type mem_t = GC | C | Big;;

type t_mach = GOI | Question | IrrevKriv of mem_t 
 | IrrevNeed of mem_t | IrrevValue of mem_t | GOInopt;;

let mach = ref GOI;;

let memname = function C -> " with copies"
               |  GC -> " with GC"
               |  Big -> " with full copies"
;;

let machinename = function
        GOI -> "GOI - call-by-name machine"
      | GOInopt -> "GOI - call-by-name - not optimized"
      | Question -> "Question stack - call-by-name machine"
      | IrrevKriv(m) -> "Irreversible - call-by-name machine (Krivine machine)"
                          ^ (memname m)
      | IrrevValue(m) -> "Irreversible - call-by-value machine"
                          ^ (memname m)
      | IrrevNeed(m) -> "Irreversible - call-by-need machine"
                          ^ (memname m)
;;

(* Limits *)

let meml = ref 5000;;
let timel = ref 2000;;

let verbose = ref true;;

type t_type = 
   { type_desc : t_type_desc;
     mutable type_level : int }

and

t_var = {mutable l : typelink}

and

t_type_desc =
   TVar of t_var
 | TArrow of t_type * t_type
 | TConst of string
 | TList of t_type

and

typelink = 
   TypConst of t_type  (* Instanciated type variable *)
 | TypVar              (* Non-instanciated type variable *)
;;

exception Error of sexpr * string;;

exception DeclError of location * string;;
exception TypeError of location * string * t_type * t_type;;

let current_level = ref 0;;

let monom = 0;;
let polym = -1;;

let new_type_var () =
  { type_desc = TVar {l = TypVar};
    type_level = !current_level }
;;

let oc = ref stdout;;

let numexpr = ref 0;;

let label_count = ref 0;;

let new_var_code x = 
   incr label_count;
   x ^ "_" ^ (string_of_int !label_count)
;;

let new_label() = new_var_code "_lbl";;

let op_type op inputt outputt lbl = 
   (op, {type_desc = TArrow(
                  {type_desc = TConst inputt;
                   type_level = monom},
                  {type_desc = TArrow(
                     {type_desc = TConst inputt;
                      type_level = monom},
                     {type_desc = TConst outputt;
                      type_level = monom});
                   type_level = monom});
                type_level = monom},
         Func("x",Func("y",Op(op, Var "x",Var "y"))), lbl);;

let f_type op inputt outputt lbl =
    (op, {type_desc = TArrow(
                     {type_desc = TConst inputt;
		      type_level = monom},
                     {type_desc = TConst outputt;
                      type_level = monom});
                   type_level = monom},
            Func("x",F(op, Var "x")),
            lbl);;

let glob = ref [
    "tt",{type_desc = TConst "bool";
          type_level = monom}, Var "tt","";
    "ff",{type_desc = TConst "bool";
          type_level = monom}, Var "ff","";
    f_type "succ" "int" "int" "_succ";
    f_type "pred" "int" "int" "_pred";
    f_type "not" "bool" "bool" "_not";
    f_type "iszero" "int" "bool" "_iszero";
    op_type "+" "int" "int" "_add";
    op_type "-" "int" "int" "_sub";
    op_type "*" "int" "int" "_pro";
    op_type "/" "int" "int" "_quo";
    op_type "%" "int" "int" "_rem";
    op_type "=" "int" "bool" "_eq";
    op_type "<" "int" "bool" "_less";
    op_type "<=" "int" "bool" "_lesseq";
    op_type "<>" "int" "bool" "_diff";
    op_type "xor" "bool" "bool" "_xor";
    op_type "and" "bool" "bool" "_and";
    op_type "or" "bool" "bool" "_or";
    "cond",(let cond_var = new_type_var() in
               cond_var.type_level <- polym;
                     {type_desc = TArrow(
                         {type_desc = TConst "bool";
		          type_level = monom},
                         {type_desc = TArrow(
                            cond_var,
			    {type_desc = TArrow(
                               cond_var,
			       cond_var);
			     type_level = polym});
                          type_level = polym});
                       type_level = polym}),
             Func("b",Func("tte",Func("ffe",Cond(Var "b",
                Var "tte",Var "ffe")))), "_cond";
    "Y", (let rec_var = new_type_var() in
             rec_var.type_level <- polym;
                    {type_desc = TArrow(
                       {type_desc = TArrow(
                          rec_var,
                          rec_var);
                        type_level = polym},
                       rec_var);
                     type_level = polym}),
                Func("x",Rec("y", Apply(Var "x", Var "y"))), "_Y";
    "::", (let cons_var = new_type_var() in
              cons_var.type_level <- polym;
                     {type_desc = TArrow(
                        cons_var,
                        {type_desc = TArrow(
                            {type_desc = TList(cons_var);
                             type_level = polym},
                            {type_desc = TList(cons_var);
                             type_level = polym});
                         type_level = polym});
                      type_level = polym}),
         Func("x",Func("l",Op("::", Var "x", Var "l"))), "_cons";
    "[]", (let nil_var = new_type_var() in
              nil_var.type_level <- polym;
              {type_desc = TList(nil_var);
                      type_level = polym}), Var "[]", "_nil";
    "hd", (let hd_var = new_type_var() in
              hd_var.type_level <- polym;
                    {type_desc = TArrow(
                        {type_desc = TList(hd_var);
                         type_level = polym},
                        hd_var);
                      type_level = polym}),
              Func("x", F("hd", Var "x")), "_hd";
    "tl", (let tl_var = new_type_var() in
              tl_var.type_level <- polym;
                     {type_desc = TArrow(
                        {type_desc = TList(tl_var);
                         type_level = polym},
                        {type_desc = TList(tl_var);
                         type_level = polym});
                      type_level = polym}),
              Func("x", F("tl", Var "x")), "_tl";
    "isnil", (let nil_var = new_type_var() in
              nil_var.type_level <- polym;
                     {type_desc = TArrow(
                        {type_desc = TList(nil_var);
                         type_level = polym},
                        {type_desc = TConst "bool";
                         type_level = monom});
                      type_level = polym}), 
              Func("x", F("isnil", Var "x")),"_isnil";
   "_match_error", (let mat_var = new_type_var() in
                    mat_var.type_level <- polym;
                    mat_var), Var "_match_error","_match_error"]
;;

(* link between priorities and the grammar
       expr => 0
       sexpr => 1
       simpleexprlist => 2
       simpleexpr => p >= 3 (several priorities because there are 
                             operators of different precedences)
*)

let output_prog highlight = 
let rec print_expr prio e0 = 
  let s = begin
  match e0 with
    Func(id,e) ->
      (if prio>1 then "(" else "") ^ "fn " ^ id ^ " => " ^ (print_expr 0 e) ^
      (if prio>1 then ")" else "")
  | Var id -> 
      id
  | Int n ->
      string_of_int n
  | Apply(e1,e2) ->
      (if (prio=1) or (prio>2) then "(" else "") ^ 
      (print_expr 2 e1) ^ " " ^ 
      (print_expr (if prio=2 then 3 else 1) e2) ^
      (if (prio=1) or (prio>2) then ")" else "")
  | Affectin(id,e1,e2) ->
      (if prio>1 then "(" else "") ^ "let " ^ id ^ " = " ^ 
      (print_expr 0 e1) ^ " in\n" ^ 
      (print_expr 0 e2) ^ (if prio>1 then ")" else "")
  | Cond(be,tte,ffe) ->
      (if prio>1 then "(" else "") ^ "if " ^ 
      (print_expr 0 be) ^ " then " ^ 
      (print_expr 0 tte) ^ " else " ^
      (print_expr 0 ffe) ^ 
      (if prio>1  then ")" else "")
  | Rec(id,e) ->
      (if prio>1 then "(" else "") ^ "rec " ^ id ^ " => " ^ 
      (print_expr 0 e) ^ 
      (if prio>1 then ")" else "")
  | Op(op,e1,e2) ->
      if op = "::" then
      begin
         (if prio>1 then "(" else "") ^ 
         (print_expr 3 e1) ^ " :: " ^
         (print_expr 0 e2) ^
         (if prio>1 then ")" else "")
      end
      else let p = match op with
           "<" | "<=" | "=" | "<>" -> 6
         | "+" | "-" | "or" | "xor" -> 8
         | "*" | "/" | "%" | "and" -> 10
         | _ -> failwith "Internal error : unknown operator"
      in (if prio>p then "(" else "") ^
         (print_expr p e1) ^ " " ^ op ^ " " ^ 
         (print_expr (p+1) e2) ^
         (if prio>p then ")" else "")
  | F(f,e) ->
      (if (prio=1) or (prio>2) then "(" else "") ^ f ^ " " ^ 
      (print_expr (if prio=2 then 3 else 1) e) ^
      (if (prio=1) or (prio>2) then ")" else "")
  end in
  if e0 == highlight then "{{ " ^ s ^ " }}" else s
        
in
function
    Affect(a,b) -> 
      "let " ^ a ^ " = " ^ (print_expr 0 b)
  | Expr(a) ->
      print_expr 0 a
;;

let print_prog oc p = 
   output_string oc ((output_prog (Var "") p) ^ "\n")
;;

let string_of_prog p = 
   output_prog (Var "") p
;;

let print_underline p e =
   output_file();
   print_endline (match p with
       Affect(_,_) -> "Error in declaration :"
     | Expr _ -> "Error in expression :");
   print_endline (output_prog e p)
;;

