(* SFL lexer definition *)

{

open Parser;;
open Lexing;;

(* For nested comments *)

let comment_depth = ref 0;;

(* The table of keywords *)

let keyword_table = [
  "let", LET;
  "fn", FUN;
  "in", IN;
  "rec", REC;
  "match", MATCH;
  "with", WITH;
  "if", IF;
  "then", THEN;
  "else", ELSE;
  "and", AND;
  "or", OR;
  "xor", XOR];;

let predef_table = ["isnil"; "hd"; "tl"; "cond"; "Y"; "tt"; "ff"; "iszero";
  "pred"; "succ"; "not"];;

type lexerr = Unterminated_comment | Illegal_character
;;

exception Lex_error of lexerr * int * int
;;

}

rule main = parse
    [' ' '\010' '\013' '\009' '\012']+
        { main lexbuf }
  | ['A'-'Z' 'a'-'z']
    ( '_' ? ['A'-'Z' 'a'-'z' '0'-'9' ] ) *
        { let s = lexeme lexbuf in
            try
              List.assoc s keyword_table 
            with Not_found ->
              if List.mem s predef_table then PREDEF s else IDENT s }
  | ['0'-'9']+
    | '0' ['x' 'X'] ['0'-'9' 'A'-'F' 'a'-'f']+
    | '0' ['o' 'O'] ['0'-'7']+
    | '0' ['b' 'B'] ['0'-'1']+
         { INT (int_of_string(lexeme lexbuf)) }
  | "(*"
      { comment_depth := 1; comment lexbuf; main lexbuf }
  | "=>" { ARROW }
  | "[" { LBRACKET }
  | "]" { RBRACKET }
  | "," { COMMA }
  | ":" { COLON }
  | "'" { QUOTE }
  | "=" { EQUAL }
  | "_" { UNDERSCORE }
  | "<" { LESS }
  | ">" { MORE }
  | "(" { LPAREN }
  | ")" { RPAREN }
  | ";" { SEMICOLON }
  | ";;" { SEMISEMI }
  | "+" { PLUS }
  | "-" { MINUS }
  | "*" { TIMES }
  | "/" { DIVIDE }
  | "%" { REMAINDER }
  | "::" { CONS }
  | "|" { BAR }
  | eof { EOF }
  | "<=" { LESSEQUAL }
  | ">=" { MOREEQUAL }
  | "<>" { DIFF }
  | "#" { SHARP }
  | '"' [ ' ' '!' '#'-'\127' ] * '"'
      { let s = lexeme lexbuf in
           STRING(String.sub s 1 (String.length s - 2)) }
  | _
      { raise (Lex_error(Illegal_character,
                            lexeme_start lexbuf, lexeme_end lexbuf)) }

and comment = parse
    "(*"
      { comment_depth := succ !comment_depth; comment lexbuf }
  | "*)"
      { comment_depth := pred !comment_depth;
        if !comment_depth > 0 then comment lexbuf }
  | eof
      { raise (Lex_error(Unterminated_comment, -1, -1)) }
  | _
      { comment lexbuf }


