open Globals;;
open Location;;

let idcount = ref 0;;

let new_ident () =
   incr idcount;
   "_id" ^ (string_of_int !idcount)
;;

let allfailexpr = Var "_match_error";;

let rec nbr x = function
       Rec(v,e) -> if x = v then 0 else (nbr x e)
    |  Op(_,e1,e2) -> (nbr x e1) + (nbr x e2)
    |  F(_,e) -> nbr x e
    |  Cond(be, tte, ffe) -> (nbr x be) + (nbr x tte) + (nbr x ffe) 
    |  Func(v,e) -> if x = v then 0 else (nbr x e)
    |  Apply(e1,e2) -> (nbr x e1) + (nbr x e2)
    |  Var v -> if v = x then 1 else 0
    |  Int _ -> 0
    |  Affectin(v,e1,e2) -> (nbr x e1) + (if x = v then 0 else nbr x e2)
;;

(* pres x e tells if the variable x is free in expression e.
   It is a bit quicker than (nbr x e>0) since it does not have
   to go through the whole expression if x is found at the beginning *)

let rec pres x = function
       Rec(v,e) -> (x<>v) & (pres x e)
    |  Op(_,e1,e2) -> (pres x e1) or (pres x e2)
    |  F(_,e) -> pres x e
    |  Cond(be,tte,ffe) -> (pres x be) or (pres x tte) or (pres x ffe) 
    |  Func(v,e) -> (x<>v) & (pres x e)
    |  Apply(e1,e2) -> (pres x e1) or (pres x e2)
    |  Var v -> v = x
    |  Int _ -> false
    |  Affectin(v,e1,e2) -> (pres x e1) or ((x <> v) & (pres x e2))
;;

(* rename is linear if v2 does not appear in e, so the internal calls
   to let v'=new_ident() in ...rename v v' e... are linear
   and rename in general is quadratic (in the size of e),
   for example in : rename a x (fn x => fn x => .... fn x => a)      *)

let rec rename v1 v2 e = 
   let rec ren = function
      Affectin(v,e1,e2) -> 
         if v1 = v then
            Affectin(v, ren e1, e2)
         else if v = v2 then
            let v' = new_ident() in
                Affectin(v', ren e1, ren(rename v v' e2))
         else
            Affectin(v, ren e1, ren e2)
    | Apply(e1,e2) -> Apply(ren e1, ren e2)
    | Func(v,e) -> 
         if v = v1 then
             Func(v,e)
         else if v = v2 then
             let v' = new_ident() in
                Func(v', ren(rename v v' e))
         else
              Func(v, ren e)
    | Var v -> if v = v1 then Var v2 else Var v
    | Int i -> Int i
    | Rec(v,e) -> 
         if v = v1 then
             Rec(v,e)
         else if v = v2 then
             let v' = new_ident() in
                Rec(v', ren(rename v v' e))
         else
             Rec(v, ren e)
    | Op(x,e1,e2) -> Op(x, ren e1, ren e2)
    | F(x,e) -> F(x, ren e)
    | Cond(be, tte, ffe) -> Cond(ren be, ren tte, ren ffe)
   in ren e
;;

(* Caution ! This function is optimized to treat only the case
when x is used at most ONCE in e2 : nbr x e2 <= 1 and
if x is not used, it raises exception Not_found *)

let rec subst x e1 e2 =
   let rec sub = function
       Int i -> raise Not_found
    |  Var v -> if v = x then e1 else raise Not_found
    |  Apply(e1',e2') -> 
          (try
             Apply(sub e1', e2')
          with Not_found ->
             Apply(e1', sub e2'))
    |  Func(v, e) ->
          if v<>x then
             if pres v e1 then
                (* Pb if v appears in e1 and x in e : variable capture *)
                let v' = new_ident() in
                   Func(v', sub(rename v v' e))
             else
                Func(v, sub e)
          else
             raise Not_found 
    |  Affectin(v,e1',e2') ->
          (try
             Affectin(v,sub e1', e2') 
          with Not_found -> if v=x then 
             raise Not_found
          else if pres v e1 then
             (* Pb if v appears in e1 and x in e2' : variable capture *)
             let v' = new_ident() in
                Affectin(v', e1', sub(rename v v' e2'))
          else
             Affectin(v, e1', sub e2') )
    |  Rec(v, e) ->
          if v<>x then
             if pres v e1 then
                (* Pb if v appears in e1 and x in e : variable capture *)
                let v' = new_ident() in
                   Rec(v', sub(rename v v' e))
             else
                Rec(v, sub e)
          else
             raise Not_found
    |  Op(v, e1', e2') -> 
          (try
             Op(v, sub e1', e2')
          with Not_found ->
             Op(v, e1', sub e2'))
    |  F(v, e) -> F(v, sub e)
    |  Cond(be, tte, ffe) ->
          (try
             Cond(sub be, tte, ffe)
          with Not_found -> try
             Cond(be, sub tte, ffe)
          with Not_found ->
             Cond(be, tte, sub ffe))
in sub e2;;

let makeaffect(x,e1,e2) =
  match nbr x e2 with
    0 -> e2
  | 1 -> subst x e1 e2
  | _ -> match e1 with 
             Var z -> rename x z e2
           | _ -> Affectin(x,e1,e2)
;;

let rec linear v p = match p.pdesc with
   Id x -> if not (List.mem x v) then x::v else
           begin
               output_location p.ploc;
               print_endline "Non-linear pattern-matching";
               raise Bad_end
           end
 | Any -> v
 | Nil -> v
 | Cons(a,b) -> linear (linear v a) b
 | In _ -> v
;;

let rec lookforopt = function
   Apply(Func(x,e1),e2) -> succeedopt (Affectin(x,e2,e1))
 | Affectin(x,e1,e2) -> begin
     match nbr x e2 with
        0 -> e2
      | 1 -> subst x e1 e2
      | _ -> 
     match e1 with 
        Var z -> rename x z e2
      | _ -> 
     try
        Affectin(x,lookforopt e1,e2)
     with
        Not_found ->
        Affectin(x,e1,lookforopt e2)
   end
 | Rec(x,e) -> if pres x e then
                   Rec(x, lookforopt e)
               else
                   succeedopt e
 | F(x,e) -> F(x, lookforopt e)
 | Op(x,e1,e2) -> begin
     try
        Op(x,lookforopt e1, e2)
     with
        Not_found -> Op(x, e1, lookforopt e2)
   end
 | Cond(be,tte,ffe) -> begin
     try
        Cond(lookforopt be, tte, ffe)
     with
        Not_found -> 
     try
        Cond(be, lookforopt tte, ffe)
     with
        Not_found -> Cond(be, tte, lookforopt ffe)
   end
 | Apply(e1,e2) -> begin
     try
        Apply(lookforopt e1, e2)
     with
        Not_found -> Apply(e1, lookforopt e2)
   end
 | Var _ | Int _ -> raise Not_found
 | Func(x,e) -> if (x = "_") or (pres x e) then
                   Func(x, lookforopt e)
                else
                   Func("_", succeedopt e)

and

succeedopt e =
  try
    lookforopt e
  with
    Not_found -> e
;;

let rec optimize e =
   try
      optimize (lookforopt e)
   with
      Not_found -> e
;;

(* There are three cases : 
     Nil       
     Cons(.,.)
     All = all the patterns after the first pattern which is not
           Nil nor Cons, even the case of failure of the whole matching
           (expression stored in efail in the code below)
Schema
  if isnil x then (Nil case) else
     let y = hd x in
     let z = tl x in
        (Cons case) 

If there is no Nil case/Cons case, it is replaced by the All case
If the Nil/Cons cases fail, the All case is called

Pattern-matching is not exhaustive iff _match_error is called
Some cases in the matching are unused iff a failexpr other than
_match_error isn't called
*)

let rec scanpat y z = function 
   [] -> ([],[],[])
 | (a::l) -> 
      match fst(a) with 
         ((_,{pdesc = Nil})::li) -> 
                   let (nil,cons,all) = scanpat y z l in 
                         ((li,snd(a))::nil,cons,all)
      |  ((_,{pdesc = Cons(b,c)})::li) -> 
                   let (nil,cons,all) = scanpat y z l in
                         (nil,((y,b)::(z,c)::li,snd(a))::cons,all)
      |  ((_,{pdesc = Any})::_) | ((_,{pdesc = Id _})::_) -> 
                   ([],[],(a::l))
      |  ((_,{pdesc = In _; ploc = loc})::l) -> 
                   output_location loc;
                   print_endline "Integers mixed in a list matching";
                   raise Bad_end
      | _ -> failwith "Internal error in pattern matching"
;;

let rec pat_match failexpr = function 
   [] -> failexpr
 | (p1::pl) as patlist -> 
match p1 with
   ((s,{pdesc = Id x})::l,e) -> pat_match (pat_match failexpr pl) 
                      [l,rename x s e] 
 | ((s,{pdesc = Any})::l,e) -> pat_match (pat_match failexpr pl) [l,e] 
 | ((s,{pdesc = Nil})::_,_) -> list_pattern s patlist failexpr
 | ((s,{pdesc = Cons(_,_)})::_,_) -> list_pattern s patlist failexpr
 | ((s,{pdesc = In i})::l,e) -> let efail = new_ident() in
                      makeaffect(efail, pat_match failexpr pl,
                      Cond(Op("=",Var s,Int i), 
                      pat_match (Var efail) [l,e],Var efail))
 | ([],e) -> e

and

list_pattern x patlist failexpr =
   let efail = new_ident() 
   and y = new_ident()
   and z = new_ident() in
   let (nil,cons,all) = scanpat y z patlist in
      makeaffect(efail, pat_match failexpr all, 
      Cond(F("isnil", Var x),
      pat_match (Var efail) nil, makeaffect(y, F("hd", Var x),
      makeaffect(z, F("tl", Var x), pat_match (Var efail) cons))))
;;

let rec assocident x = List.map (fun (pat,e) -> ([x,pat],z_to_s e))

and

z_to_s e0 = match e0.desc with
   ZFunc(lpat) -> 
    List.iter (fun (pat,_) -> (linear [] pat);()) lpat;
    let x = new_ident() in
       let patexpr = pat_match allfailexpr (assocident x lpat) in
          if pres "_match_error" patexpr then
          begin
             output_location e0.loc;
             print_string "Warning : pattern-matching not exhaustive\n"
          end;
          Func(x,patexpr)
 | ZVar x -> Var x
 | ZInt i -> Int i
 | ZApply(e1, e2) ->
     begin
       let e = z_to_s e2 in
       match e1.desc with
          ZVar x -> 
            begin
             match x with
                "Y" -> let x = new_ident() in
                         Rec(x, Apply(e, Var x))
              | "cond" -> let tte = new_ident()
                          and ffe = new_ident()
                          in Func(tte, Func(ffe, Cond(e, 
                               Var tte, Var ffe)))
              | "hd" | "tl" | "isnil" | "succ" | "pred" | "iszero" ->
                          F(x, e)
              | _ -> Apply(Var x, e)
            end
        | _ -> Apply(z_to_s e1, e) 
     end
 | ZAffectin(x,e1,e2) -> Affectin(x, z_to_s e1, z_to_s e2)
 | ZCond(be,tte,ffe) -> Cond(z_to_s be, z_to_s tte, z_to_s ffe)
 | ZRec(x,e) -> Rec(x, z_to_s e)
 | ZOp(x,e1,e2) -> Op(x, z_to_s e1, z_to_s e2)
 | ZMatch(e,lpat) ->    
    List.iter (fun (pat,_) -> (linear [] pat);()) lpat;
    let patexpr =
       match e.desc with
          ZVar s -> pat_match allfailexpr (assocident s lpat)
        | _ -> 
            let x = new_ident() in
              makeaffect(x, z_to_s e, pat_match allfailexpr (assocident x lpat))
    in
       if pres "_match_error" patexpr then
       begin
           output_location e0.loc;
           print_string "Warning : pattern-matching not exhaustive\n"
       end;
       patexpr
 | ZType(e,typeexpr) ->
    z_to_s e
;;

let optimizeprog = function
   ZAffect(x,e) -> Affect(x, optimize (z_to_s e))
 | ZExpr e -> Expr (optimize (z_to_s e))
;;
