(* Test mode *)

open Globals;;
open Typing;;
open Pattern;;
open Location;;
open Compiler;;

type t_info = Found of int | NotFound | NotTested;;

let rec insertl v = function [] -> [v]
                 |  (a::l) -> if a=v then a::l else (a::insertl v l) 
;;

let addelem l v =
   l := insertl v !l
;;

let rec posspace s n = 
   if n = String.length s then n else
   if String.get s n = ' ' then n else
   posspace s (n+1)
;;

let rec index x = function [] -> raise Not_found
                  | (a::l) -> if x = a then 0 else 1+index x l
;; 

let endmsg = " - SFL program running...";;

let prectime = "Precise user time ";;
let rep = "(Repeated ";;
let numrep = " times) ";;

let infocmd = 
  ["maxmem"; "totmem"; "time"; "copy"; "others"; "!"; "d"; "r"; "s"; "p"; "q";
   "t"; "h"; "l"; "i"; "beta"; "cont"];;

let infotext = 
  ["Maximum memory requirement : ";
   "Total memory requirement : ";
   "User time : ";
   "Number of copied cells : ";
   "Number of printings, arithmetic and other operations : ";
   "Number of ! and !* : ";
   "Number of d and d* : ";
   "Number of r and r* : ";
   "Number of s and s* : ";
   "Number of p and p* : ";
   "Number of q and q* : ";
   "Number of t and t* : ";
   "Number of h and h* : ";
   "Number of l and l* : ";
   "Number of i and i* : "]
;;

(* You can translate the strings in the two following arrays
   to get your TeX documents in the language you want *)

let infoname =
  ["Memory requirement (bytes)";
   "Total memory (bytes)";
   "Time (ms)";
   "Number of copied cells";
   "Number of other operations";
   "Number of ! and !* : ";
   "Number of d and d* : ";
   "Number of r and r* : ";
   "Number of s and s* : ";
   "Number of p and p* : ";
   "Number of q and q* : ";
   "Number of t and t* : ";
   "Number of h and h* : ";
   "Number of l and l* : ";
   "Number of i and i* : ";
   "$\\beta$-reductions";
   "Contractions"]
;;

let machname =
  [GOI, "GOI";
   Question, "Question";
   GOInopt, "GOI non-opt";
   IrrevKriv(Big), "Krivine (fc)";
   IrrevKriv(C), "Krivine (c)";
   IrrevKriv(GC), "Krivine (gc)";
   IrrevValue(Big), "Value (fc)";
   IrrevValue(C), "Value (c)";
   IrrevValue(GC), "Value (gc)";
   IrrevNeed(GC), "Need"]
;;

let proglist = ref ([]: (string * t_type * sexpr * string) list);;
let machlist = ref ([]: t_mach list);;
let perffile = ref "a.perf";;
let texfile = ref "a.tex";;
let infolist = ref ([] : int list);;

let help() =
   print_string ("
SFL toplevel by Bruno Blanchet 1995
Test mode

Commands :

    #canceltest;; : Exit test mode
    #runtest;; : Run the test
    #gettest;; : Create TeX output from an existing performance file
    #info INFONAME ...;; : Add information data request
        where INFONAME may be
          maxmem, totmem, time, beta, copy, !, d, r, s, cont,
          t, p, q, h, l, i, others
        Present requested informations are :");
   List.iter (fun i -> print_string (" " ^ (List.nth infocmd i))) !infolist;
   print_string ("   
    #texfile \"TexFile\";; : Set tex output file (now " ^ !texfile ^ ")
    #perffile \"PerfFile\";; : Set performance file (now " ^ !perffile ^ ")
    #help;; #h;; : displays this help screen
    #goi;; : activates the GOI machine
    #goi nopt;; : activates the not optimized GOI machine
    #quest;; : activates the question stack machine
    #irrev MACH MEM;; : activates an irreversible machine
       where MACH can be :
          kriv : Krivine machine (call-by-name)
          need : call-by-need machine
          value : call-by-value machine
       and MEM can be :
          gc : with garbage collector
          c : copy environments 
          fc : full copy of environments
    Selected machines are : \n");
   List.iter (fun m -> print_endline ("      " ^ machinename m)) !machlist;
   print_newline()
;;

let runone onexpr =
                   let outputc = open_out !outname in
                      oc := outputc;
                      (try
                         machine (onexpr::!glob)
                      with
                         x -> close_out outputc;
                              raise x);
                      close_out outputc;
                      if !exename = "" then exename := "./a.out";
                      print_string "Compiling C file ...";
                      print_newline();
                      if Sys.command (Config.gcc ^ " -o " ^ !exename ^ 
                         " " ^ !outname) = 0 then
                         if Sys.command (!exename ^ " | tee a.perf.tmp") = 0 then ()
                         else
                         begin
                            print_endline "Execution failed";
                            failwith "Program failure"
                         end
                      else
                      begin
                         print_endline "C compilation failed. This should \
                            never happen";
                         failwith "Compilation failure"
                      end
;;

let getinfo datarec = 
   let ch_perf = open_in "a.perf.tmp" in
   try
      while true do
         let line = input_line ch_perf in
         let index = ref 0 in
         List.iter (fun s ->
            let leng = String.length s in
            if (String.length line > leng) & (String.sub line 0 leng = s) then
               datarec.(!index) <- Found (int_of_string (String.sub line leng
                 ((posspace line leng) - leng )));
            incr index) infotext
      done
   with
      End_of_file -> close_in ch_perf
;;

let runtest data = 
     latex := false;
     let machindex = ref 0 in
     List.iter (fun m ->
       mach := m;
       data.(!machindex) <- Array.create 
         (List.length !proglist) (Array.create 16 NotTested);
       (try
         let progindex = ref 0 in
         List.iter (fun onexpr ->
            count := true;
            repeat := 1;
            runone onexpr;
            (* Extrait les informations du fichier de performances *)
            Sys.command ("cat a.perf.tmp >> " ^ !perffile);
            data.(!machindex).(!progindex) <- Array.create 16 NotFound;
            getinfo data.(!machindex).(!progindex);
            count := false;
            (match data.(!machindex).(!progindex).(2) with
                  Found time ->
                     if time < 20 then
                        repeat := 1000
                     else if time < 200 then
                        repeat := 100
                     else if time < 2000 then
                        repeat := 10
                     else
                        repeat := 1
                 | _ -> repeat := 1);
            runone onexpr;
            (* Extrait le temps d'execution du fichier de performances *)
            data.(!machindex).(!progindex).(15) <- Found !repeat; 
            getinfo data.(!machindex).(!progindex);
            (match data.(!machindex).(!progindex).(2) with
                  Found time ->
                     let ch_perf = open_out_gen [Open_wronly; Open_append; Open_text] (0o644)
                        !perffile in
                     output_string ch_perf prectime;
                     if !repeat <> 1 then
                        output_string ch_perf (rep ^ (string_of_int !repeat)
                          ^ numrep);
                     output_string ch_perf (": " ^ (string_of_float 
                       (float time /. float !repeat)) ^ " ms\n");
                     close_out ch_perf
                | _ -> ());
            incr progindex
          ) !proglist
       with Failure _ -> ());
       incr machindex
     ) !machlist
;;

let gettest data =
     for i=0 to (List.length !machlist)-1 do
        data.(i) <- Array.create (List.length !proglist) (Array.create 16 NotTested)
     done;
     let machindex = ref (-1) in
     let progindex = ref 0 in
     let ch_perf = open_in !perffile in
       (try
          while true do
             let line = input_line ch_perf in
                if !machindex <> -1 then
                begin
                   let iindex = ref 0 in
                   List.iter (fun s ->
                      if !iindex <> 2 then (* Don't capture time the same way *)
                      begin
                         let leng = String.length s in
                         if (String.length line > leng) & (String.sub line 0 
                         leng = s) then
                            data.(!machindex).(!progindex).(!iindex) <- 
                             Found (int_of_string (String.sub 
                             line leng ((posspace line leng) - leng )))
                      end
                      else
                      begin
                         let leng = String.length prectime in    
                         let repleng = String.length rep in
                         if (String.length line > leng) & (String.sub line 0 
                         leng = prectime) then
                         begin
                            if (String.length line > repleng + leng) & 
                            (String.sub line leng repleng = rep) then
                            begin
                               let endrep = posspace line (repleng+leng) in
                               let repeatnum = (int_of_string (String.sub     
                                 line (repleng+leng) (endrep-repleng-leng))) in
                               data.(!machindex).(!progindex).(15) <- 
                                 Found repeatnum;
                               data.(!machindex).(!progindex).(2) <-
                                 Found (truncate ((float_of_string
                                 (String.sub line (endrep+10) (String.length line - endrep - 10))) *. float repeatnum))
                            end
                            else
                            begin
                               data.(!machindex).(!progindex).(15) <- Found 1;
                               data.(!machindex).(!progindex).(2) <- 
                                 Found (truncate (float_of_string (String.sub     
                                 line (leng+2) ((posspace line (leng+2)) - (leng+2)))))
                            end
                         end
                      end;
                      incr iindex) infotext
                end;
                if (String.length line > String.length endmsg) & 
                  (String.sub line (String.length line - String.length 
                  endmsg) (String.length endmsg) = endmsg) then
                   machindex := -1;
                List.iter (fun m ->
                   if line = ((machinename m) ^ endmsg) then
                      let line = input_line ch_perf in
                         List.iter (function (_,_,e,_) as onexpr ->
                            if line = string_of_prog (Expr e) then
                            begin
                               machindex := index m !machlist;
                               progindex := index onexpr !proglist;
                               data.(!machindex).(!progindex) <- 
                                  Array.create 16 NotFound
                            end) !proglist
                   ) !machlist
          done
       with End_of_file -> ());
       close_in ch_perf
;;

let createtex data =
   let ch_tex = open_out_gen [Open_wronly; Open_append; Open_text; Open_creat] 0o644 !texfile in
      List.iter (fun i ->
          output_string ch_tex "\\vbox{\\begin{tabular}{|>{\\tt}l|";
          if i = 2 then
             List.iter (fun _ -> output_string ch_tex "D{.}{.}{3}|") !machlist
          else
             List.iter (fun _ -> output_string ch_tex "r|") !machlist;
          output_string ch_tex ("}\n\\hline\n\\multicolumn{1}{|l|}{" ^ 
               (List.nth infoname i) ^"}");
          List.iter (fun m -> output_string ch_tex ("& \\multicolumn{1}{c|}{" ^
               (List.assoc m machname) ^"}")) !machlist;
          output_string ch_tex "\\cr\n\\hline\n";
          let progindex = ref 0 in
          List.iter (fun (_,_,e,_) -> 
             output_string ch_tex (string_of_prog (Expr e));
             let machindex = ref 0 in
             let get i = match data.(!machindex).(!progindex).(i) with
                 Found x -> x
               | NotFound -> 0
               | NotTested -> raise Not_found
             in
             List.iter (fun m -> 
                output_string ch_tex ("& " ^
                try
                   if i = 2 then
                   begin
                       if get 15 = 0 then "\\multicolumn{1}{c|}{Not found}" 
                       else
                       string_of_float ((float (get 2)) /. (float (get 15)))
                   end
                   else if i = 15 then
                       match m with
                          GOI | GOInopt -> string_of_int ((get 10) / 4)
                        | Question -> string_of_int ((get 10) / 2)
                        | _ -> string_of_int (get 6)
                   else if i = 16 then
                       string_of_int ((get 7) + (get 8))
                   else
                       string_of_int (get i)
                with Not_found -> if i = 2 then
                      "\\multicolumn{1}{c|}{Not tested}"
                   else "Not tested");
                incr machindex
             ) !machlist;
             output_string ch_tex "\\cr\n";
             incr progindex
          ) !proglist;
          output_string ch_tex "\\hline\n\\end{tabular}\n}\\kern 7mm\n\n"
      ) !infolist;
      close_out ch_tex
;;

let test lexbuf =
let fin = ref 0
and oldmach = !mach
in proglist := [];
   machlist := [];
   perffile := "a.perf";
   texfile := "a.tex";
   infolist := [];
let rec skip() =
  try
    match Lexer.main lexbuf with
       Parser.EOF | Parser.SEMISEMI -> ()
     | _ -> skip()
  with Lexer.Lex_error(_,_,_) ->
      skip() in
let skip_maybe() =
  if Parsing.is_current_lookahead Parser.EOF
  or Parsing.is_current_lookahead Parser.SEMISEMI
  then () else skip() in 
  act_input := Toplevel lexbuf;
  globalsym := makegsym !glob;
while !fin = 0 do
   print_string "#";
   flush stdout;
   try
      idcount := 0;
      match Parser.top Lexer.main lexbuf with
         Prog p ->
            reset_type_vars();
            let ty = type_prog p in
            let p2 = optimizeprog p in
            (match p2 with
                Affect(x,e) ->
                   print_string "No let declaration in test mode"
              | Expr e ->
                   addelem proglist ("_expr_0",ty,e,"_expr_0"))
       | Command(c,p) ->
           match c with
                   "canceltest" -> fin := 1
                 | "runtest" -> fin := 2
                 | "gettest" -> fin := 3
                 | "help" | "h" -> help()
                 | "info" -> List.iter (function (PId s) -> 
                        (try
                            addelem infolist (index s infocmd)
                         with
                            Not_found -> print_endline ("Illegal parameter " ^ s ^ " for #info"))
                      |  _ -> print_endline "#info can \
                             only have identifiers as parameters") p
                 | "texfile" -> (match p with
                        [PStr s] -> texfile:=s
                      | _ -> print_endline "#texfile can only have a \
                                  filename as parameter")
                 | "perffile" -> (match p with
                       [PStr s] -> perffile:=s
                      | _ -> print_endline "#perffile can only have a \
                                  filename as parameter")
                 | "goi" -> (match p with
                         [] -> addelem machlist GOI
                       | [PId "nopt"] -> addelem machlist GOInopt
                       | _ -> print_endline "#goi can only have nopt as \
                                 parameter.")
                 | "quest" -> addelem machlist Question
                 | "irrev" ->
                      let m = ref GC in
                      let machtmp = ref (IrrevKriv GC) in
                      List.iter (
                       function  PId "kriv" -> machtmp := IrrevKriv(GC)
                       | PId "need" -> machtmp := IrrevNeed(GC)
                       | PId "value" -> machtmp := IrrevValue(GC)
                       | PId "gc" -> m := GC
                       | PId "c" -> m := C
                       | PId "fc" -> m := Big
                       | _ -> print_endline "#irrev can only have kriv, need \
                                 or value as parameter.") p;
                      (match !machtmp with
                          IrrevKriv(_) -> addelem machlist (IrrevKriv(!m))
                        | IrrevValue(_) -> addelem machlist (IrrevValue(!m))
                        | _ -> 
                             if !m = GC then
                                addelem machlist (IrrevNeed(GC))
                             else
                                print_endline "The call-by-need machine \
                                    cannot copy environments"
                       )
                 | _ -> print_endline ("Unknown command #"^c^". Ignored")
   with Failure s -> print_endline s
   | Bad_end -> ()
   | Sys_error s -> print_endline ("Input/output error : " ^ s)
   | Parsing.Parse_error ->
       let loc = (Lexing.lexeme_start lexbuf,
                  Lexing.lexeme_end lexbuf) in
       skip_maybe();
       output_location loc;
       print_endline "Syntax error."
   | Lexer.Lex_error (errcode, pos1, pos2) ->
       skip();
       (match errcode with
           Lexer.Unterminated_comment -> 
             output_file();
             print_endline "Lexical error. Unterminated comment."
         | Lexer.Illegal_character ->
             output_location (pos1, pos2);
             print_endline "Lexical error. Illegal character.")
   | _ -> print_endline "Unknown exception"
done;
if !fin >= 2 then
begin
   let data = Array.create (List.length !machlist) (Array.create
      (List.length !proglist) (Array.create 16 NotTested)) in 
   if !fin = 2 then runtest data;
   if !fin = 3 then gettest data;
   if !infolist <> [] then createtex data
end;
mach := oldmach
;;
