(* Main program for the toplevel *)

open Globals;;
open Typing;;
open Pattern;;
open Location;;
open Compiler;;

let load filename =
   let inputc = open_in_bin filename in
   let lexbuf = Lexing.from_channel inputc in
   act_input := File(filename, inputc);
   globalsym := makegsym !glob;
   try
     while true do
       match anal_phrase lexbuf with
          (ty,Affect(x,e)) -> 
             glob := (x, ty, e, x) :: !glob
        | _ -> if !verbose then 
                  print_string "Warning : Evaluable expression ignored\n" 
     done
   with End_of_file ->
     close_in inputc
   | x ->
     close_in inputc;
     raise x
;;
   
let help()=
   print_string ("
SFL toplevel by Bruno Blanchet 1995

Commands :

    #out \"OutputFile\";; (now " ^ !outname ^ ")
    #exe \"ExeFile\";; (now : " ^ (if !exename = "" then "a.out"
        else !exename) ^ ")
    #typing;;/#notyping;; : activates/desactivates typing
       (" ^ (if !types then "activated" else "unactivated") ^ " now)
    #latex;;/#nolatex;; : produce code with Latex trace or not
       present state is #" ^ (if !latex then "" else "no") ^ "latex 
    #count;;/#nocount;; : count operations on running or not
       present state is #" ^ (if !count then "" else "no") ^ "count 
    #verbose;; : display information messages or not (" ^ 
       (if !verbose then "yes" else "not") ^ " now)
    #repeat N;; : repeats evaluation N times (now " ^ (string_of_int !repeat)
     ^ ")
    #goi;; : activates the GOI machine
    #goi nopt;; : activates the not optimized GOI machine
    #quest;; : activates the question stack machine
    #irrev MACH MEM;; : activates an irreversible machine
       where MACH can be :
          kriv : Krivine machine (call-by-name)
          need : call-by-need machine
          value : call-by-value machine
       and MEM can be :
          gc : with garbage collector
          c : copy environments 
          fc : full copy of environments
    Present machine is " ^ (machinename !mach) ^ "
    #load \"File\";; : loads a program (with no evaluable expressions)
           The identifiers defined in this program are kept forever in
           memory for use by other programs
    #run \"File\";; : runs a program on the selected machine
    #compile \"File\";; : compiles the program
    #test;; : Enter test mode
    #tex;; : compiles and shows the TeX trace
    #limit mem N;; : limits data size to N kbytes (now " ^ 
      (string_of_int !meml) ^ ")
    #limit time N;; : limits CPU time to N seconds (now " ^
      (string_of_int !timel) ^ ")
    #help;; #h;; : displays this help screen
    #quit;; #exit;; : exits the toplevel
\n")
;;

let print_machine() =
   print_endline (machinename !mach)
;;

let runfile f =
    if !exename = "" then exename := "./a.out";
    secure compile f;
    Sys.command !exename; 
    ()
;;

let main() =
if !verbose then help();
let lexbuf = Lexing.from_channel stdin in
let rec skip() =
  try
    match Lexer.main lexbuf with
       Parser.EOF | Parser.SEMISEMI -> ()
     | _ -> skip()
  with Lexer.Lex_error(_,_,_) ->
      skip() in
let skip_maybe() =
  if Parsing.is_current_lookahead Parser.EOF
  or Parsing.is_current_lookahead Parser.SEMISEMI
  then () else skip() in 
while true do
   act_input := Toplevel lexbuf;
   globalsym := makegsym !glob;
   print_string "#";
   flush stdout;
   try
      idcount := 0;
      match Parser.top Lexer.main lexbuf with
         Prog p ->
            reset_type_vars();
            let ty = type_prog p in
            let p2 = optimizeprog p in
            if !verbose then print_prog stdout p2;
            (match p2 with
                Affect(x,e) ->
                   glob := (x, ty, e, x) :: !glob
              | Expr e ->
                      oc := open_out !outname;
                      (try
                         machine (("_expr_0", ty, e, "_expr_0") :: !glob)
                      with
                         x -> close_out !oc;
                              raise x);
                      close_out !oc;
                      if !exename = "" then exename := "./a.out";
                      print_string "Compiling C file ...";
                      print_newline();
                      if Sys.command (Config.gcc ^ " -o " ^ !exename ^ 
                         " " ^ !outname) = 0 then
                         (Sys.command !exename; ())
                      else
                         print_endline "C compilation failed. This should \
                            never happen")
       |  Command(c,p) ->
           match c with
                   "quit" | "exit" -> exit 0
                 | "out" -> (match p with
                         [PStr s] -> outname := s;
                                     print_endline ("C output directed to " ^ s)
                       | _ -> print_endline "#out can only have a filename \
                                  as parameter")
                 | "exe" -> (match p with
                         [PStr s] -> exename := s;
                                     print_endline ("Executable output directed to " ^ s)
                       | _ -> print_endline "#exe can only have a filename \
                                  as parameter")
                 | "typing" -> types := true
                 | "notyping" -> types := false
                 | "help" | "h" -> help()
                 | "latex" -> latex := true
                 | "nolatex" -> latex := false
                 | "count" -> count := true
                 | "nocount" -> count := false
                 | "repeat" -> (match p with
                          [PInt n] -> if n>0 then 
                                         repeat:=n
                                      else
                                         print_endline "Number of evaluations must be at least 1"
                        | _ -> print_endline "#repeat must be followed by a positive integer")
                 | "goi" -> (match p with
                         [] -> mach := GOI; print_machine()
                       | [PId "nopt"] -> mach := GOInopt; print_machine()
                       | _ -> print_endline "#goi can only have nopt as \
                                 parameter.")
                 | "quest" -> mach := Question; print_machine()
                 | "irrev" ->
                      let m = ref GC in
                      let machtmp = ref (IrrevKriv GC) in
                      List.iter (
                       function  PId "kriv" -> machtmp := IrrevKriv(GC)
                       | PId "need" -> machtmp := IrrevNeed(GC)
                       | PId "value" -> machtmp := IrrevValue(GC)
                       | PId "gc" -> m := GC
                       | PId "c" -> m := C
                       | PId "fc" -> m := Big
                       | _ -> print_endline "#irrev can only have kriv, need \
                                 or value as parameter.") p;
                      (match !machtmp with
                          IrrevKriv(_) -> mach := IrrevKriv(!m); print_machine()
                        | IrrevValue(_) -> mach := IrrevValue(!m); print_machine()
                        | _ -> 
                             if !m = GC then
                             begin
                                mach := IrrevNeed(!m); 
                                print_machine()
                             end
                             else
                                print_endline "The call-by-need machine \
                                    cannot copy environments"
                       )
                 | "load" -> (match p with
                         [PStr s] -> secure load s
                       | _ -> print_endline "#load can only have a \
                                 filename as parameter")
                 | "run" -> (match p with
                         [PStr s] -> runfile s
                       | _ -> print_endline "#run can only have a \
                                 filename as parameter")
                 | "test" -> print_endline "Entering test mode";
                             Test.test lexbuf;
                             print_endline "Leaving test mode"
                 | "compile" -> (match p with
                           [PStr s] -> if !exename = "" then exename := "./a.out";
                                       secure compile s
                         | _ -> print_endline "#compile can only have a \
                                  filename as parameter")
                 | "tex" -> (try
                               close_in (open_in_bin "./a.tex");
                               if Sys.command (Config.latex ^ " ./a.tex") = 0 then
                                   (Sys.command (Config.xdvi ^ " ./a.dvi &"); ())
                               else
                                   print_endline "Latex compilation failed"
                            with
                               Sys_error _ ->
                                 print_endline "There is no Tex trace")
                 | "limit" -> (match p with
                          [PId "mem"; PInt n] -> meml := n
                        | [PId "time"; PInt n] -> timel := n
                        | _ -> print_endline "#limit should have mem N or \
                                  time N as arguments")
                 | "verbose" -> verbose := not !verbose
                 | _ -> print_endline ("Unknown command #"^c^". Ignored")
 with Failure s -> print_endline s
   | Bad_end -> ()
   | Sys_error s -> print_endline ("Input/output error : " ^ s)
   | Parsing.Parse_error ->
       let loc = (Lexing.lexeme_start lexbuf,
                  Lexing.lexeme_end lexbuf) in
       skip_maybe();
       output_location loc;
       print_endline "Syntax error."
   | Lexer.Lex_error (errcode, pos1, pos2) ->
       skip();
       (match errcode with
           Lexer.Unterminated_comment -> 
             output_file();
             print_endline "Lexical error. Unterminated comment."
         | Lexer.Illegal_character ->
             output_location (pos1, pos2);
             print_endline "Lexical error. Illegal character.")
   | End_of_file -> exit 0
   | _ -> print_endline "Unknown exception"
done
;;
