(* Typing of expressions *)
(* Algorithm used = destructive unification *)

open Globals;;
open Location;;

(* Computes a type from a type expression *)

let typeexpr_vars = ref ([] : (string * t_type) list);;

let reset_typeexpr_vars() = 
    typeexpr_vars := []
;;

let rec type_of_typeexpr t = match t.tdesc with
   ZTypeVar v -> 
      begin
         try
            List.assoc v !typeexpr_vars
         with
            Not_found ->
               let tvar = new_type_var() in
                  typeexpr_vars := (v, tvar) :: !typeexpr_vars;
                  tvar
      end
 | ZTypeArrow(t1,t2) ->
      { type_desc = TArrow(type_of_typeexpr t1, type_of_typeexpr t2);
        type_level = !current_level }
 | ZTypeConstr(t1,s) ->
      if s = "list" then
         { type_desc = TList(type_of_typeexpr t1);
           type_level = !current_level }
      else
      begin
         output_location t.tloc;
         print_endline "Type constructor list only.";
         raise Bad_end
      end
 | ZTypeConst s ->
      if (s = "int") or (s = "bool") then
         { type_desc = TConst s;
           type_level = monom }
      else
      begin
         output_location t.tloc;
         print_endline "Simple types int and bool only.";
         raise Bad_end
      end
;;   

(* Computes the value of a type variable, with path compression *)

let rec instanvar t = match t.type_desc with  
   TVar ({l = TypConst ty } as link) -> let ty2 = instanvar ty in
            link.l <- TypConst ty2; ty2
 | _ -> t
;;

(* Generalizes a type *)

let rec generalize t = let t = instanvar t in
 begin
 match t.type_desc with
   TVar {l = TypVar} -> if t.type_level > !current_level then t.type_level <- polym
 | TVar {l = TypConst _} -> failwith "Internal error : there should not be \
                                  an instanciated variable here"
 | TArrow(t1,t2) -> let lvl1 = generalize t1
                    and lvl2 = generalize t2
                    in
                    t.type_level <- if lvl1 <= lvl2 then 
                       lvl1
                    else
                       lvl2
 | TConst _ -> t.type_level <- monom
 | TList(t1) -> t.type_level <- generalize t1
 end;
 t.type_level
;;

(* Unification *)

exception Unify_err;;

let rec occur_check lev v t= let t = instanvar t in
 match t.type_desc with
   TArrow(t1,t2) -> 
      occur_check lev v t1;
      occur_check lev v t2
 | TConst _ -> ()
 | TList t1 -> occur_check lev v t1
 | TVar {l = TypVar} -> 
      if v==t then 
         raise Unify_err
      else if t.type_level > lev then
         t.type_level <- lev 
 | TVar {l = TypConst ty} -> occur_check lev v ty
;;

let rec unify ty1 ty2 =
 if ty1 == ty2 then () else
 let ty1 = instanvar ty1
 and ty2 = instanvar ty2 
 in if ty1 == ty2 then () else
 match ty1.type_desc, ty2.type_desc with
   (TVar ({l = TypVar} as tlink1)), (TVar ({l = TypVar} as tlink2)) ->
       if ty1.type_level < ty2.type_level then
       begin
          ty2.type_level <- ty1.type_level;
          tlink2.l <- TypConst ty1
       end else begin
          ty1.type_level <- ty2.type_level;
          tlink1.l <- TypConst ty2
       end
 | (TVar ({l = TypVar} as tlink)), _ ->
       occur_check ty1.type_level ty1 ty2; 
       tlink.l <- TypConst ty2
 | _, (TVar ({l = TypVar} as tlink)) -> 
       occur_check ty2.type_level ty2 ty1;
       tlink.l <- TypConst ty1
 | TArrow (ta1,tb1), TArrow (ta2, tb2) -> unify ta1 ta2;
                                         unify tb1 tb2
 | TConst s1, TConst s2 -> if s1<>s2 then raise Unify_err
 | TList t1, TList t2 -> unify t1 t2
 | _ -> raise Unify_err
;;

(* Type instances *)

let rec type_copy t = 
  if t.type_level = polym then
     match t.type_desc with
        TVar ({l = TypVar} as link) -> 
           (let v = new_type_var() in link.l <- TypConst v; v )
      | TVar {l = TypConst ty} -> ty
      | TArrow(t1,t2) -> {type_desc = TArrow(type_copy t1, type_copy t2);
                          type_level = monom}
      | TList t1 -> {type_desc = TList (type_copy t1);
                     type_level = monom}
      | TConst s -> {type_desc = TConst s; type_level = monom}
  else
     match t.type_desc with
        TVar {l = TypConst ty} -> type_copy ty
      | _ -> t 
;;

let rec type_clean t = match t.type_desc with
   TVar {l = TypVar} -> ()
 | TVar ({l = TypConst ty} as link) -> if t.type_level = polym then 
                link.l <- TypVar
            else
                type_clean ty 
 | TArrow(t1,t2) -> if t.type_level = polym then
                (type_clean t1; type_clean t2)
 | TList t1 -> if t.type_level = polym then type_clean t1
 | TConst _ -> ()
;;

let type_instance ty =
   let ty2 = type_copy ty in
     type_clean ty;
     ty2
;;

(* Typing of patterns and expressions *)

let rec type_pat typx tres symbtable lpat =
let symb = ref symbtable in
let rec type1pat p = match p.pdesc with
   Nil -> let tvar = new_type_var() in
              {type_desc = TList tvar; type_level = !current_level}
 | Cons(a,b) -> let t1 = {type_desc = TList (type1pat a);
                          type_level = !current_level}
                and t2 = type1pat b
                in (try
                      unify t1 t2
                   with
                      Unify_err ->
                         raise (TypeError(b.ploc,"Pattern",t2,t1)));
                   t2
 | Id x -> let tvar = new_type_var()
           in symb:=(x, tvar):: !symb;
              tvar
 | Any -> new_type_var()
 | In _ -> {type_desc = TConst "int"; type_level = monom}
in
List.iter (fun (pat,expr) -> 
   symb := symbtable;
   let tp1 = type1pat pat
   and tr1 = type_expr !symb expr
   in (try
         unify tres tr1
      with
         Unify_err ->
            raise (TypeError(expr.loc,"Expression",tr1,tres)));
      (try
         unify typx tp1
      with
         Unify_err ->
            raise (TypeError(pat.ploc,"Pattern",tp1,typx)))) lpat

and 

type_expr symbtable e0 = match e0.desc with
   ZOp(op,e1,e2) ->
     let t1' = type_expr symbtable e1 
     and t2' = type_expr symbtable e2 in
     (match (type_instance (List.assoc op symbtable)).type_desc with
        TArrow(t1,{type_desc = TArrow(t2,t3)}) ->
         (try
            unify t1 t1'
         with
            Unify_err ->
              raise (TypeError(e1.loc,"Expression",t1',t1)));
         (try
            unify t2 t2'
         with
            Unify_err ->
              raise (TypeError(e2.loc,"Expression",t2',t2)));
         t3
      | _ -> failwith "Internal error : bad typed operator")
 | ZRec(x,e) -> 
     let typx = new_type_var() in
     let ty = type_expr ((x,typx)::symbtable) e in
        (try 
           unify ty typx
        with
           Unify_err ->
              raise (TypeError(e.loc,"Expression",ty,typx)));
        typx
 | ZCond(be,tte,ffe) ->
     let tbe = type_expr symbtable be in
     (try
        unify {type_desc = TConst "bool"; type_level = monom} tbe
     with
        Unify_err ->
           raise (TypeError(be.loc,"Expression",tbe,{type_desc = TConst "bool";
                                    type_level = monom})));
     let ttte = type_expr symbtable tte
     and tffe = type_expr symbtable ffe
     in (try
           unify ttte tffe
        with
           Unify_err ->
              raise (TypeError(ffe.loc,"Expression",tffe,ttte)));
        tffe
 | ZMatch(e,lpat) ->
     let tye = type_expr symbtable e in
     let tres = new_type_var() in
        type_pat tye tres symbtable lpat;
        tres
 | ZFunc(lpat) -> 
     let typx = new_type_var() in
     let tres = new_type_var() in
        type_pat typx tres symbtable lpat;
        {type_desc = TArrow(typx, tres); type_level = !current_level}
 | ZVar x -> 
     (try
        type_instance (List.assoc x symbtable)
      with
        Not_found -> raise (DeclError(e0.loc,"Variable " ^ x ^ " not declared")))
 | ZInt(n) -> { type_desc = TConst "int"; type_level = monom }
 | ZApply(e1,e2) ->
     let tres = new_type_var() in
     let t1 = type_expr symbtable e1 in
     let t2 = { type_desc = TArrow(type_expr symbtable e2,tres);
                type_level = !current_level } in
        (try
            unify t1 t2
        with Unify_err -> raise (TypeError(e1.loc,"Expression",t1,t2)));
        tres
 | ZAffectin(x,e1,e2) -> 
     incr current_level;
     let type1=type_expr symbtable e1 in
        decr current_level;
        generalize type1;
           type_expr ((x,type1)::symbtable) e2
 | ZType(e,typeexpr) ->
     reset_typeexpr_vars();
     let t2 = type_of_typeexpr typeexpr in
     if !types then
     begin
        let t1 = type_expr symbtable e in
        try
           unify t1 t2;
           t2
        with
           Unify_err ->
              raise (TypeError(e.loc, "Expression", t1, t2))
     end
     else
     begin
        try
           let t1 = type_expr symbtable e in
              unify t1 t2;
              t2
        with
           Unify_err | TypeError(_,_,_,_) ->
              t2
     end
;;


let rec makegsym = 
   function [] -> []
    |  ((x, t, _, _)::l) -> (x, t)::makegsym l
;; 

let globalsym = ref (makegsym !glob);;

let type_vars = ref ([] : (t_type * string) list);;
let type_vars_counter = ref (-1);;

let reset_type_vars ()= 
   type_vars := [];
   type_vars_counter := -1
;;

let rec int_to_alpha n = if n=0 then "'V" else 
    (int_to_alpha (n / 26)) ^ (String.make 1 (Char.chr (n mod 26 + 65)));;

let new_type_var_name () =
    incr type_vars_counter;
    int_to_alpha !type_vars_counter
;;

let rec print_type t prio = let t = instanvar t in
  match t.type_desc with
    TVar {l = TypVar} -> (try
                      print_string (List.assq t !type_vars)
                   with
                      Not_found -> let s = new_type_var_name() in 
                                        type_vars := (t,s) :: !type_vars;
                                        print_string s)
 | TVar _ -> failwith "Internal error in print_type"
 | TArrow(t1,t2) -> if prio >= 1 then print_string "(";
                    print_type t1 1; 
                    print_string " => "; 
                    print_type t2 0;
                    if prio >= 1 then print_string ")"
 | TList t1 -> print_type t1 1;
               print_string " list"
 | TConst s -> print_string s
;;

let type_prog p =
  current_level := 0;
  try
     match p with
       ZAffect(x,e) ->
         incr current_level; 
         let ty =type_expr !globalsym e in
            decr current_level;
            generalize ty;
            if !verbose then
            begin
               print_string (x^" : "); 
               print_type ty 0; 
               print_newline()
            end;
            globalsym := (x, ty):: !globalsym;
            ty
     | ZExpr e -> let ty = type_expr !globalsym e in
                    if !verbose then
                    begin
                       print_type ty 0;
                       print_newline()
                    end;
                    ty
  with
    DeclError(loc,msg) ->
       output_location loc;
       print_endline msg;
       raise Bad_end
  | TypeError(loc,msg,t1,t2) ->
       output_location loc;
       print_string msg;
       print_string " of type ";
       print_type t1 0;
       print_string "\ncannot be used with type ";
       print_type t2 0;
       print_endline ".";
       if !types then
          raise Bad_end;
       print_endline "Warning : expression is considered of unknown type.";
       let tvar = new_type_var() in
          tvar.type_level <- polym;
          tvar
;;
